import logging
from datetime import timedelta
from hashlib import md5
from hashlib import sha256

try:
    from urllib import urlencode

    from urlparse import parse_qs
    from urlparse import urlsplit
    from urlparse import urlunsplit
except ImportError:
    from urllib.parse import parse_qs
    from urllib.parse import urlencode
    from urllib.parse import urlsplit
    from urllib.parse import urlunsplit
from uuid import uuid4

from django.utils import timezone

from oidc_provider import settings
from oidc_provider.lib.claims import StandardScopeClaims
from oidc_provider.lib.errors import AuthorizeError
from oidc_provider.lib.errors import ClientIdError
from oidc_provider.lib.errors import RedirectUriError
from oidc_provider.lib.utils.common import get_browser_state_or_default
from oidc_provider.lib.utils.token import create_code
from oidc_provider.lib.utils.token import create_id_token
from oidc_provider.lib.utils.token import create_token
from oidc_provider.lib.utils.token import encode_id_token
from oidc_provider.models import Client
from oidc_provider.models import UserConsent

logger = logging.getLogger(__name__)


class AuthorizeEndpoint(object):
    _allowed_prompt_params = {"none", "login", "consent", "select_account"}
    client_class = Client

    def __init__(self, request):
        self.request = request
        self.params = {}

        self._extract_params()

        # Determine which flow to use.
        if self.params["response_type"] in ["code"]:
            self.grant_type = "authorization_code"
        elif self.params["response_type"] in ["id_token", "id_token token", "token"]:
            self.grant_type = "implicit"
        elif self.params["response_type"] in ["code token", "code id_token", "code id_token token"]:
            self.grant_type = "hybrid"
        else:
            self.grant_type = None

        # Determine if it's an OpenID Authentication request (or OAuth2).
        self.is_authentication = "openid" in self.params["scope"]

    def _extract_params(self):
        """
        Get all the params used by the Authorization Code Flow
        (and also for the Implicit and Hybrid).

        See: http://openid.net/specs/openid-connect-core-1_0.html#AuthRequest
        """
        # Because in this endpoint we handle both GET
        # and POST request.
        query_dict = self.request.POST if self.request.method == "POST" else self.request.GET

        self.params["client_id"] = query_dict.get("client_id", "")
        self.params["redirect_uri"] = query_dict.get("redirect_uri", "")
        self.params["response_type"] = query_dict.get("response_type", "")
        self.params["scope"] = query_dict.get("scope", "").split()
        self.params["state"] = query_dict.get("state", "")
        self.params["nonce"] = query_dict.get("nonce", "")

        self.params["prompt"] = self._allowed_prompt_params.intersection(
            set(query_dict.get("prompt", "").split())
        )

        self.params["code_challenge"] = query_dict.get("code_challenge", "")
        self.params["code_challenge_method"] = query_dict.get("code_challenge_method", "")

    def validate_params(self):
        # Client validation.
        try:
            self.client = self.client_class.objects.get(client_id=self.params["client_id"])
        except Client.DoesNotExist:
            logger.debug("[Authorize] Invalid client identifier: %s", self.params["client_id"])
            raise ClientIdError()

        # Redirect URI validation.
        if self.is_authentication and not self.params["redirect_uri"]:
            logger.debug("[Authorize] Missing redirect uri.")
            raise RedirectUriError()
        if self.params["redirect_uri"] not in self.client.redirect_uris:
            logger.debug("[Authorize] Invalid redirect uri: %s", self.params["redirect_uri"])
            raise RedirectUriError()

        # Grant type validation.
        if not self.grant_type:
            logger.debug("[Authorize] Invalid response type: %s", self.params["response_type"])
            raise AuthorizeError(
                self.params["redirect_uri"], "unsupported_response_type", self.grant_type
            )

        if not self.is_authentication and (
            self.grant_type == "hybrid"
            or self.params["response_type"] in ["id_token", "id_token token"]
        ):
            logger.debug("[Authorize] Missing openid scope.")
            raise AuthorizeError(self.params["redirect_uri"], "invalid_scope", self.grant_type)

        # Nonce parameter validation.
        if self.is_authentication and self.grant_type == "implicit" and not self.params["nonce"]:
            raise AuthorizeError(self.params["redirect_uri"], "invalid_request", self.grant_type)

        # Response type parameter validation.
        if (
            self.is_authentication
            and self.params["response_type"] not in self.client.response_type_values()
        ):
            raise AuthorizeError(self.params["redirect_uri"], "invalid_request", self.grant_type)

        # PKCE validation of the transformation method.
        if self.params["code_challenge"]:
            if self.params["code_challenge_method"] not in ["plain", "S256"]:
                raise AuthorizeError(
                    self.params["redirect_uri"], "invalid_request", self.grant_type
                )

    def create_code(self):
        code = create_code(
            user=self.request.user,
            client=self.client,
            scope=self.params["scope"],
            nonce=self.params["nonce"],
            is_authentication=self.is_authentication,
            code_challenge=self.params["code_challenge"],
            code_challenge_method=self.params["code_challenge_method"],
        )

        return code

    def create_token(self):
        token = create_token(
            user=self.request.user,
            client=self.client,
            scope=self.params["scope"],
        )

        return token

    def create_response_uri(self):
        uri = urlsplit(self.params["redirect_uri"])
        query_params = parse_qs(uri.query)
        query_fragment = {}

        try:
            if self.grant_type in ["authorization_code", "hybrid"]:
                code = self.create_code()
                code.save()
            if self.grant_type == "authorization_code":
                query_params["code"] = code.code
                query_params["state"] = self.params["state"] if self.params["state"] else ""
            elif self.grant_type in ["implicit", "hybrid"]:
                token = self.create_token()

                # Check if response_type must include access_token in the response.
                if self.params["response_type"] in [
                    "id_token token",
                    "token",
                    "code token",
                    "code id_token token",
                ]:
                    query_fragment["access_token"] = token.access_token

                # We don't need id_token if it's an OAuth2 request.
                if self.is_authentication:
                    kwargs = {
                        "token": token,
                        "user": self.request.user,
                        "aud": self.client.client_id,
                        "nonce": self.params["nonce"],
                        "request": self.request,
                        "scope": self.params["scope"],
                    }
                    # Include at_hash when access_token is being returned.
                    if "access_token" in query_fragment:
                        kwargs["at_hash"] = token.at_hash
                    id_token_dic = create_id_token(**kwargs)

                    # Check if response_type must include id_token in the response.
                    if self.params["response_type"] in [
                        "id_token",
                        "id_token token",
                        "code id_token",
                        "code id_token token",
                    ]:
                        query_fragment["id_token"] = encode_id_token(id_token_dic, self.client)
                else:
                    id_token_dic = {}

                # Store the token.
                token.id_token = id_token_dic
                token.save()

                # Code parameter must be present if it's Hybrid Flow.
                if self.grant_type == "hybrid":
                    query_fragment["code"] = code.code

                query_fragment["token_type"] = "bearer"

                query_fragment["expires_in"] = settings.get("OIDC_TOKEN_EXPIRE")

                query_fragment["state"] = self.params["state"] if self.params["state"] else ""

            if settings.get("OIDC_SESSION_MANAGEMENT_ENABLE"):
                # Generate client origin URI from the redirect_uri param.
                redirect_uri_parsed = urlsplit(self.params["redirect_uri"])
                client_origin = "{0}://{1}".format(
                    redirect_uri_parsed.scheme, redirect_uri_parsed.netloc
                )

                # Create random salt.
                salt = md5(uuid4().hex.encode()).hexdigest()

                # The generation of suitable Session State values is based
                # on a salted cryptographic hash of Client ID, origin URL,
                # and OP browser state.
                session_state = "{client_id} {origin} {browser_state} {salt}".format(
                    client_id=self.client.client_id,
                    origin=client_origin,
                    browser_state=get_browser_state_or_default(self.request),
                    salt=salt,
                )
                session_state = sha256(session_state.encode("utf-8")).hexdigest()
                session_state += "." + salt
                if self.grant_type == "authorization_code":
                    query_params["session_state"] = session_state
                elif self.grant_type in ["implicit", "hybrid"]:
                    query_fragment["session_state"] = session_state

        except Exception as error:
            logger.exception("[Authorize] Error when trying to create response uri: %s", error)
            raise AuthorizeError(self.params["redirect_uri"], "server_error", self.grant_type)

        uri = uri._replace(
            query=urlencode(query_params, doseq=True),
            fragment=uri.fragment + urlencode(query_fragment, doseq=True),
        )

        return urlunsplit(uri)

    def set_client_user_consent(self):
        """
        Save the user consent given to a specific client.

        Return None.
        """
        date_given = timezone.now()
        expires_at = date_given + timedelta(days=settings.get("OIDC_SKIP_CONSENT_EXPIRE"))

        uc, created = UserConsent.objects.get_or_create(
            user=self.request.user,
            client=self.client,
            defaults={
                "expires_at": expires_at,
                "date_given": date_given,
            },
        )
        uc.scope = self.params["scope"]

        # Rewrite expires_at and date_given if object already exists.
        if not created:
            uc.expires_at = expires_at
            uc.date_given = date_given

        uc.save()

    def client_has_user_consent(self):
        """
        Check if already exists user consent for some client.

        Return bool.
        """
        value = False
        try:
            uc = UserConsent.objects.get(user=self.request.user, client=self.client)
            if (set(self.params["scope"]).issubset(uc.scope)) and not (uc.has_expired()):
                value = True
        except UserConsent.DoesNotExist:
            pass

        return value

    def is_client_allowed_to_skip_consent(self):
        implicit_flow_resp_types = {"id_token", "id_token token"}
        return (
            self.client.client_type != "public"
            or self.params["response_type"] in implicit_flow_resp_types
        )

    def get_scopes_information(self):
        """
        Return a list with the description of all the scopes requested.
        """
        scopes = StandardScopeClaims.get_scopes_info(self.params["scope"])
        if settings.get("OIDC_EXTRA_SCOPE_CLAIMS"):
            scopes_extra = settings.get("OIDC_EXTRA_SCOPE_CLAIMS", import_str=True).get_scopes_info(
                self.params["scope"]
            )
            for index_extra, scope_extra in enumerate(scopes_extra):
                for index, scope in enumerate(scopes[:]):
                    if scope_extra["scope"] == scope["scope"]:
                        del scopes[index]
        else:
            scopes_extra = []

        return scopes + scopes_extra



import hashlib
import logging

try:
    from urllib import urlencode

    from urlparse import parse_qs, urlsplit, urlunsplit
except ImportError:
    from urllib.parse import urlsplit, parse_qs, urlunsplit, urlencode

from Cryptodome.PublicKey import RSA
from django.contrib.auth.views import redirect_to_login

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from django.contrib.auth import logout as django_user_logout
from django.core.cache import cache
from django.db import transaction
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.template.loader import render_to_string
from django.utils.decorators import method_decorator
from django.views.decorators.clickjacking import xframe_options_exempt
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods
from django.views.generic import TemplateView, View
from jwkest import long_to_base64

from oidc_provider import settings, signals
from oidc_provider.compat import get_attr_or_callable
from oidc_provider.lib.claims import StandardScopeClaims
from oidc_provider.lib.endpoints.authorize import AuthorizeEndpoint
from oidc_provider.lib.endpoints.introspection import TokenIntrospectionEndpoint
from oidc_provider.lib.endpoints.token import TokenEndpoint
from oidc_provider.lib.errors import (
    AuthorizeError,
    ClientIdError,
    RedirectUriError,
    TokenError,
    TokenIntrospectionError,
    UserAuthError,
)
from oidc_provider.lib.utils.authorize import strip_prompt_login
from oidc_provider.lib.utils.common import (
    cors_allow_any,
    get_issuer,
    get_site_url,
    redirect,
)
from oidc_provider.lib.utils.oauth2 import protected_resource_view
from oidc_provider.lib.utils.token import client_id_from_id_token
from oidc_provider.models import Client, ResponseType, RSAKey

logger = logging.getLogger(__name__)

OIDC_TEMPLATES = settings.get("OIDC_TEMPLATES")
after_end_session_hook = settings.get("OIDC_AFTER_END_SESSION_HOOK", import_str=True)


class AuthorizeView(View):
    authorize_endpoint_class = AuthorizeEndpoint

    def get(self, request, *args, **kwargs):
        authorize = self.authorize_endpoint_class(request)

        try:
            authorize.validate_params()

            if get_attr_or_callable(request.user, "is_authenticated"):
                # Check if there's a hook setted.
                hook_resp = settings.get("OIDC_AFTER_USERLOGIN_HOOK", import_str=True)(
                    request=request, user=request.user, client=authorize.client
                )
                if hook_resp:
                    return hook_resp

                if "login" in authorize.params["prompt"]:
                    if "none" in authorize.params["prompt"]:
                        raise AuthorizeError(
                            authorize.params["redirect_uri"], "login_required", authorize.grant_type
                        )
                    else:
                        django_user_logout(request)
                        next_page = strip_prompt_login(request.get_full_path())
                        return redirect_to_login(next_page, settings.get("OIDC_LOGIN_URL"))

                if "select_account" in authorize.params["prompt"]:
                    # TODO: see how we can support multiple accounts for the end-user.
                    if "none" in authorize.params["prompt"]:
                        raise AuthorizeError(
                            authorize.params["redirect_uri"],
                            "account_selection_required",
                            authorize.grant_type,
                        )
                    else:
                        django_user_logout(request)
                        return redirect_to_login(
                            request.get_full_path(), settings.get("OIDC_LOGIN_URL")
                        )

                if {"none", "consent"}.issubset(authorize.params["prompt"]):
                    raise AuthorizeError(
                        authorize.params["redirect_uri"], "consent_required", authorize.grant_type
                    )

                if not authorize.client.require_consent and (
                    authorize.is_client_allowed_to_skip_consent()
                    and "consent" not in authorize.params["prompt"]
                ):
                    return redirect(authorize.create_response_uri())

                if authorize.client.reuse_consent:
                    # Check if user previously give consent.
                    if authorize.client_has_user_consent() and (
                        authorize.is_client_allowed_to_skip_consent()
                        and "consent" not in authorize.params["prompt"]
                    ):
                        return redirect(authorize.create_response_uri())

                if "none" in authorize.params["prompt"]:
                    raise AuthorizeError(
                        authorize.params["redirect_uri"], "consent_required", authorize.grant_type
                    )

                # Generate hidden inputs for the form.
                context = {
                    "params": authorize.params,
                }
                hidden_inputs = render_to_string("oidc_provider/hidden_inputs.html", context)

                # Remove `openid` from scope list
                # since we don't need to print it.
                if "openid" in authorize.params["scope"]:
                    authorize.params["scope"].remove("openid")

                context = {
                    "client": authorize.client,
                    "hidden_inputs": hidden_inputs,
                    "params": authorize.params,
                    "scopes": authorize.get_scopes_information(),
                }

                return render(request, OIDC_TEMPLATES["authorize"], context)
            else:
                if "none" in authorize.params["prompt"]:
                    raise AuthorizeError(
                        authorize.params["redirect_uri"], "login_required", authorize.grant_type
                    )
                if "login" in authorize.params["prompt"]:
                    next_page = strip_prompt_login(request.get_full_path())
                    return redirect_to_login(next_page, settings.get("OIDC_LOGIN_URL"))

                return redirect_to_login(request.get_full_path(), settings.get("OIDC_LOGIN_URL"))

        except (ClientIdError, RedirectUriError) as error:
            context = {
                "error": error.error,
                "description": error.description,
            }

            return render(request, OIDC_TEMPLATES["error"], context)

        except AuthorizeError as error:
            uri = error.create_uri(authorize.params["redirect_uri"], authorize.params["state"])

            return redirect(uri)

    def post(self, request, *args, **kwargs):
        authorize = self.authorize_endpoint_class(request)

        try:
            authorize.validate_params()

            if not request.POST.get("allow"):
                signals.user_decline_consent.send(
                    self.__class__,
                    user=request.user,
                    client=authorize.client,
                    scope=authorize.params["scope"],
                )

                raise AuthorizeError(
                    authorize.params["redirect_uri"], "access_denied", authorize.grant_type
                )

            signals.user_accept_consent.send(
                self.__class__,
                user=request.user,
                client=authorize.client,
                scope=authorize.params["scope"],
            )

            # Save the user consent given to the client.
            authorize.set_client_user_consent()

            uri = authorize.create_response_uri()

            return redirect(uri)

        except AuthorizeError as error:
            uri = error.create_uri(authorize.params["redirect_uri"], authorize.params["state"])

            return redirect(uri)


class TokenView(View):
    token_endpoint_class = TokenEndpoint

    def post(self, request, *args, **kwargs):
        token = self.token_endpoint_class(request)

        try:
            with transaction.atomic():
                token.validate_params()
                dic = token.create_response_dic()

            return self.token_endpoint_class.response(dic)

        except TokenError as error:
            return self.token_endpoint_class.response(error.create_dict(), status=400)
        except UserAuthError as error:
            return self.token_endpoint_class.response(error.create_dict(), status=403)


@require_http_methods(["GET", "POST", "OPTIONS"])
@protected_resource_view(["openid"])
def userinfo(request, *args, **kwargs):
    """
    Create a dictionary with all the requested claims about the End-User.
    See: http://openid.net/specs/openid-connect-core-1_0.html#UserInfoResponse

    Return a dictionary.
    """

    def set_headers(response):
        response["Cache-Control"] = "no-store"
        response["Pragma"] = "no-cache"
        cors_allow_any(request, response)
        return response

    if request.method == "OPTIONS":
        return set_headers(HttpResponse())

    token = kwargs["token"]

    dic = {
        "sub": token.id_token.get("sub"),
    }

    standard_claims = StandardScopeClaims(token)
    dic.update(standard_claims.create_response_dic())

    if settings.get("OIDC_EXTRA_SCOPE_CLAIMS"):
        extra_claims = settings.get("OIDC_EXTRA_SCOPE_CLAIMS", import_str=True)(token)
        dic.update(extra_claims.create_response_dic())

    success_response = JsonResponse(dic, status=200)
    set_headers(success_response)

    return success_response


class ProviderInfoView(View):
    _types_supported = None

    @property
    def types_supported(self):
        if self._types_supported is None:
            self._types_supported = [
                response_type.value for response_type in ResponseType.objects.all()
            ]
        return self._types_supported

    def _build_response_dict(self, request):
        dic = dict()

        site_url = get_site_url(request=request)
        dic["issuer"] = get_issuer(site_url=site_url, request=request)

        dic["authorization_endpoint"] = site_url + reverse("oidc_provider:authorize")
        dic["token_endpoint"] = site_url + reverse("oidc_provider:token")
        dic["userinfo_endpoint"] = site_url + reverse("oidc_provider:userinfo")
        dic["end_session_endpoint"] = site_url + reverse("oidc_provider:end-session")
        dic["introspection_endpoint"] = site_url + reverse("oidc_provider:token-introspection")

        dic["response_types_supported"] = self.types_supported

        dic["jwks_uri"] = site_url + reverse("oidc_provider:jwks")

        dic["id_token_signing_alg_values_supported"] = ["HS256", "RS256"]

        # See: http://openid.net/specs/openid-connect-core-1_0.html#SubjectIDTypes
        dic["subject_types_supported"] = ["public"]

        dic["token_endpoint_auth_methods_supported"] = ["client_secret_post", "client_secret_basic"]

        if settings.get("OIDC_SESSION_MANAGEMENT_ENABLE"):
            dic["check_session_iframe"] = site_url + reverse("oidc_provider:check-session-iframe")

        return dic

    def _build_cache_key(self, request):
        """
        Cache key will be a combination of site URL and types supported by the provider.
        """
        key_data = get_site_url(request=request) + "".join(self.types_supported)
        key_hash = hashlib.md5(key_data.encode("utf-8")).hexdigest()
        return f"oidc_discovery_{key_hash}"

    def get(self, request):
        if settings.get("OIDC_DISCOVERY_CACHE_ENABLE"):
            cache_key = self._build_cache_key(request)
            cached_dict = cache.get(cache_key)
            if cached_dict:
                response_dict = cached_dict
            else:
                response_dict = self._build_response_dict(request)
                cache.set(cache_key, response_dict, settings.get("OIDC_DISCOVERY_CACHE_EXPIRE"))
        else:
            response_dict = self._build_response_dict(request)

        response = JsonResponse(response_dict)
        response["Access-Control-Allow-Origin"] = "*"

        return response


class JwksView(View):
    def get(self, request, *args, **kwargs):
        dic = dict(keys=[])

        for rsakey in RSAKey.objects.all():
            public_key = RSA.importKey(rsakey.key).publickey()
            dic["keys"].append(
                {
                    "kty": "RSA",
                    "alg": "RS256",
                    "use": "sig",
                    "kid": rsakey.kid,
                    "n": long_to_base64(public_key.n),
                    "e": long_to_base64(public_key.e),
                }
            )

        response = JsonResponse(dic)
        response["Access-Control-Allow-Origin"] = "*"

        return response


class EndSessionView(View):
    http_method_names = ["get", "post"]

    @classmethod
    def logout_user(
        cls,
        request,
        id_token_hint=None,
        post_logout_redirect_uri=None,
        state=None,
        client=None,
        next_page=None,
    ):
        after_end_session_hook(
            request=request,
            id_token=id_token_hint,
            post_logout_redirect_uri=post_logout_redirect_uri,
            state=state,
            client=client,
            next_page=next_page,
        )
        django_user_logout(request)

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        self.id_token_hint = request.POST.get("id_token_hint") or request.GET.get("id_token_hint")
        self.post_logout_redirect_uri = request.POST.get(
            "post_logout_redirect_uri"
        ) or request.GET.get("post_logout_redirect_uri")
        self.state = request.POST.get("state") or request.GET.get("state")
        self.client = None

        if self.id_token_hint:
            client_id = client_id_from_id_token(self.id_token_hint)
            try:
                self.client = Client.objects.get(client_id=client_id)

                if self.post_logout_redirect_uri:
                    if self.post_logout_redirect_uri not in self.client.post_logout_redirect_uris:
                        return redirect(
                            reverse("oidc_provider:end-session-prompt")
                            + "?"
                            + urlencode(
                                {
                                    "client_id": client_id,
                                }
                            )
                        )
                elif self.client.post_logout_redirect_uris:
                    self.post_logout_redirect_uri = self.client.post_logout_redirect_uris[0]
                else:
                    self.logout_user(
                        request,
                        self.id_token_hint,
                        self.post_logout_redirect_uri,
                        self.state,
                        self.client,
                    )
                    return render(
                        request, "oidc_provider/end_session_completed.html", {"client": self.client}
                    )

                if self.state:
                    uri = urlsplit(self.post_logout_redirect_uri)
                    query_params = parse_qs(uri.query)
                    query_params["state"] = self.state
                    uri = uri._replace(query=urlencode(query_params, doseq=True))
                    next_page = urlunsplit(uri)
                else:
                    next_page = self.post_logout_redirect_uri

                self.logout_user(
                    request,
                    self.id_token_hint,
                    self.post_logout_redirect_uri,
                    self.state,
                    self.client,
                    next_page,
                )
                return redirect(next_page)
            except Client.DoesNotExist:
                pass

        return redirect(reverse("oidc_provider:end-session-prompt"))


class EndSessionPromptView(TemplateView):
    http_method_names = ["get", "post"]
    template_name = "oidc_provider/end_session_prompt.html"

    def dispatch(self, request, *args, **kwargs):
        self.client_id = request.GET.get("client_id")
        self.client = Client.objects.filter(client_id=self.client_id).first()
        return super(EndSessionPromptView, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        # If user is not authenticated, we should redirect to client post logout uri if exists,
        # otherwhise, just raise a not found error.
        if not get_attr_or_callable(request.user, "is_authenticated"):
            if self.client and self.client.post_logout_redirect_uris:
                return redirect(self.client.post_logout_redirect_uris[0])
            else:
                return render(
                    request, "oidc_provider/end_session_completed.html", {"client": self.client}
                )

        return super(EndSessionPromptView, self).get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(EndSessionPromptView, self).get_context_data(**kwargs)
        context["client"] = self.client

        end_session_prompt_url = reverse("oidc_provider:end-session-prompt")
        if self.client_id:
            end_session_prompt_url += "?" + urlencode(
                {
                    "client_id": self.client_id,
                }
            )
        context["end_session_prompt_url"] = end_session_prompt_url

        return context

    def post(self, request, *args, **kwargs):
        allowed = request.POST.get("allow")
        next_page = (
            self.client.post_logout_redirect_uris[0]
            if self.client and self.client.post_logout_redirect_uris
            else None
        )

        # Only logout users if they allow it.
        if allowed:
            EndSessionView.logout_user(request, client=self.client, next_page=next_page)

        # Redirect to post logout uri if client is present.
        if next_page:
            return redirect(next_page)
        elif allowed:
            return render(
                request, "oidc_provider/end_session_completed.html", {"client": self.client}
            )
        else:
            return render(request, "oidc_provider/end_session_failed.html", {"client": self.client})


class CheckSessionIframeView(View):
    @method_decorator(xframe_options_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super(CheckSessionIframeView, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        return render(request, "oidc_provider/check_session_iframe.html", kwargs)


class TokenIntrospectionView(View):
    token_instrospection_endpoint_class = TokenIntrospectionEndpoint

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super(TokenIntrospectionView, self).dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        introspection = self.token_instrospection_endpoint_class(request)

        try:
            introspection.validate_params()
            dic = introspection.create_response_dic()
            return self.token_instrospection_endpoint_class.response(dic)
        except TokenIntrospectionError:
            return self.token_instrospection_endpoint_class.response({"active": False})



from datetime import timedelta
import time
import uuid

from Cryptodome.PublicKey.RSA import importKey
from django.utils import dateformat, timezone
from jwkest.jwk import RSAKey as jwk_RSAKey
from jwkest.jwk import SYMKey
from jwkest.jws import JWS
from jwkest.jwt import JWT

from oidc_provider.lib.utils.common import get_issuer, run_processing_hook
from oidc_provider.lib.claims import StandardScopeClaims
from oidc_provider.models import (
    Code,
    RSAKey,
    Token,
)
from oidc_provider import settings


def create_id_token(token, user, aud, nonce="", at_hash="", request=None, scope=None):
    """
    Creates the id_token dictionary.
    See: http://openid.net/specs/openid-connect-core-1_0.html#IDToken
    Return a dic.
    """
    if scope is None:
        scope = []
    sub = settings.get("OIDC_IDTOKEN_SUB_GENERATOR", import_str=True)(user=user)

    expires_in = settings.get("OIDC_IDTOKEN_EXPIRE")

    # Convert datetimes into timestamps.
    now = int(time.time())
    iat_time = now
    exp_time = int(now + expires_in)
    user_auth_time = user.last_login or user.date_joined
    auth_time = int(dateformat.format(user_auth_time, "U"))

    dic = {
        "iss": get_issuer(request=request),
        "sub": sub,
        "aud": str(aud),
        "exp": exp_time,
        "iat": iat_time,
        "auth_time": auth_time,
    }

    if nonce:
        dic["nonce"] = str(nonce)

    if at_hash:
        dic["at_hash"] = at_hash

    # Inlude (or not) user standard claims in the id_token.
    if settings.get("OIDC_IDTOKEN_INCLUDE_CLAIMS"):
        standard_claims = StandardScopeClaims(token)
        dic.update(standard_claims.create_response_dic())

        if settings.get("OIDC_EXTRA_SCOPE_CLAIMS"):
            extra_claims = settings.get("OIDC_EXTRA_SCOPE_CLAIMS", import_str=True)(token)
            dic.update(extra_claims.create_response_dic())

    dic = run_processing_hook(
        dic, "OIDC_IDTOKEN_PROCESSING_HOOK", user=user, token=token, request=request
    )

    return dic


def encode_id_token(payload, client):
    """
    Represent the ID Token as a JSON Web Token (JWT).
    Return a hash.
    """
    keys = get_client_alg_keys(client)
    _jws = JWS(payload, alg=client.jwt_alg)
    return _jws.sign_compact(keys)


def decode_id_token(token, client):
    """
    Represent the ID Token as a JSON Web Token (JWT).
    Return a hash.
    """
    keys = get_client_alg_keys(client)
    return JWS().verify_compact(token, keys=keys)


def client_id_from_id_token(id_token):
    """
    Extracts the client id from a JSON Web Token (JWT).
    Returns a string or None.
    """
    payload = JWT().unpack(id_token).payload()
    aud = payload.get("aud", None)
    if aud is None:
        return None
    if isinstance(aud, list):
        return aud[0]
    return aud


def create_token(user, client, scope, id_token_dic=None):
    """
    Create and populate a Token object.
    Return a Token object.
    """
    token = Token()
    token.user = user
    token.client = client
    token.access_token = uuid.uuid4().hex

    if id_token_dic is not None:
        token.id_token = id_token_dic

    token.refresh_token = uuid.uuid4().hex
    token.expires_at = timezone.now() + timedelta(seconds=settings.get("OIDC_TOKEN_EXPIRE"))
    token.scope = scope

    return token


def create_code(
    user, client, scope, nonce, is_authentication, code_challenge=None, code_challenge_method=None
):
    """
    Create and populate a Code object.
    Return a Code object.
    """
    code = Code()
    code.user = user
    code.client = client

    code.code = uuid.uuid4().hex

    if code_challenge and code_challenge_method:
        code.code_challenge = code_challenge
        code.code_challenge_method = code_challenge_method

    code.expires_at = timezone.now() + timedelta(seconds=settings.get("OIDC_CODE_EXPIRE"))
    code.scope = scope
    code.nonce = nonce
    code.is_authentication = is_authentication

    return code


def get_client_alg_keys(client):
    """
    Takes a client and returns the set of keys associated with it.
    Returns a list of keys.
    """
    if client.jwt_alg == "RS256":
        keys = []
        for rsakey in RSAKey.objects.all():
            keys.append(jwk_RSAKey(key=importKey(rsakey.key), kid=rsakey.kid))
        if not keys:
            raise Exception("You must add at least one RSA Key.")
    elif client.jwt_alg == "HS256":
        keys = [SYMKey(key=client.client_secret, alg=client.jwt_alg)]
    else:
        raise Exception("Unsupported key algorithm.")

    return keys



import hashlib
import inspect
import logging
from base64 import urlsafe_b64encode

from django.contrib.auth import authenticate
from django.db import DatabaseError
from django.http import JsonResponse

from oidc_provider import settings
from oidc_provider.lib.errors import TokenError
from oidc_provider.lib.errors import UserAuthError
from oidc_provider.lib.utils.oauth2 import extract_client_auth
from oidc_provider.lib.utils.token import create_id_token
from oidc_provider.lib.utils.token import create_token
from oidc_provider.lib.utils.token import encode_id_token
from oidc_provider.models import Client
from oidc_provider.models import Code
from oidc_provider.models import Token

logger = logging.getLogger(__name__)


class TokenEndpoint(object):
    def __init__(self, request):
        self.request = request
        self.params = {}
        self.user = None
        self._extract_params()

    def _extract_params(self):
        client_id, client_secret = extract_client_auth(self.request)

        self.params["client_id"] = client_id
        self.params["client_secret"] = client_secret
        self.params["redirect_uri"] = self.request.POST.get("redirect_uri", "")
        self.params["grant_type"] = self.request.POST.get("grant_type", "")
        self.params["code"] = self.request.POST.get("code", "")
        self.params["state"] = self.request.POST.get("state", "")
        self.params["scope"] = self.request.POST.get("scope", "")
        self.params["refresh_token"] = self.request.POST.get("refresh_token", "")
        # PKCE parameter.
        self.params["code_verifier"] = self.request.POST.get("code_verifier")

        self.params["username"] = self.request.POST.get("username", "")
        self.params["password"] = self.request.POST.get("password", "")

    def validate_params(self):
        try:
            self.client = Client.objects.get(client_id=self.params["client_id"])
        except Client.DoesNotExist:
            logger.debug("[Token] Client does not exist: %s", self.params["client_id"])
            raise TokenError("invalid_client")

        if self.client.client_type == "confidential":
            if not (self.client.client_secret == self.params["client_secret"]):
                logger.debug(
                    "[Token] Invalid client secret: client %s do not have secret %s",
                    self.client.client_id,
                    self.client.client_secret,
                )
                raise TokenError("invalid_client")

        if self.params["grant_type"] == "authorization_code":
            if self.params["redirect_uri"] not in self.client.redirect_uris:
                logger.debug("[Token] Invalid redirect uri: %s", self.params["redirect_uri"])
                raise TokenError("invalid_client")

            try:
                self.code = Code.objects.select_for_update(nowait=True).get(
                    code=self.params["code"]
                )
            except DatabaseError:
                logger.debug("[Token] Code cannot be reused: %s", self.params["code"])
                raise TokenError("invalid_grant")
            except Code.DoesNotExist:
                logger.debug("[Token] Code does not exist: %s", self.params["code"])
                raise TokenError("invalid_grant")

            if not (self.code.client == self.client) or self.code.has_expired():
                logger.debug("[Token] Invalid code: invalid client or code has expired")
                raise TokenError("invalid_grant")

            # Validate PKCE parameters.
            if self.code.code_challenge:
                if self.params["code_verifier"] is None:
                    raise TokenError("invalid_grant")

                if self.code.code_challenge_method == "S256":
                    new_code_challenge = (
                        urlsafe_b64encode(
                            hashlib.sha256(self.params["code_verifier"].encode("ascii")).digest()
                        )
                        .decode("utf-8")
                        .replace("=", "")
                    )
                else:
                    new_code_challenge = self.params["code_verifier"]

                # TODO: We should explain the error.
                if not (new_code_challenge == self.code.code_challenge):
                    raise TokenError("invalid_grant")

        elif self.params["grant_type"] == "password":
            if not settings.get("OIDC_GRANT_TYPE_PASSWORD_ENABLE"):
                raise TokenError("unsupported_grant_type")

            auth_args = (self.request,)
            try:
                inspect.getcallargs(authenticate, *auth_args)
            except TypeError:
                auth_args = ()

            user = authenticate(
                *auth_args, username=self.params["username"], password=self.params["password"]
            )

            if not user:
                raise UserAuthError()

            self.user = user

        elif self.params["grant_type"] == "refresh_token":
            if not self.params["refresh_token"]:
                logger.debug("[Token] Missing refresh token")
                raise TokenError("invalid_grant")

            try:
                self.token = Token.objects.get(
                    refresh_token=self.params["refresh_token"], client=self.client
                )

            except Token.DoesNotExist:
                logger.debug(
                    "[Token] Refresh token does not exist: %s", self.params["refresh_token"]
                )
                raise TokenError("invalid_grant")
        elif self.params["grant_type"] == "client_credentials":
            if not self.client._scope:
                logger.debug("[Token] Client using client credentials with empty scope")
                raise TokenError("invalid_scope")
        else:
            logger.debug("[Token] Invalid grant type: %s", self.params["grant_type"])
            raise TokenError("unsupported_grant_type")

    def validate_requested_scopes(self):
        """
        Handling validation of requested scope for grant_type=[password|client_credentials]
        """
        token_scopes = []
        if self.params["scope"]:
            # See https://tools.ietf.org/html/rfc6749#section-3.3
            # The value of the scope parameter is expressed
            # as a list of space-delimited, case-sensitive strings
            for scope_requested in self.params["scope"].split(" "):
                if scope_requested in self.client.scope:
                    token_scopes.append(scope_requested)
                else:
                    logger.debug(
                        "[Token] The request scope %s is not supported by client %s",
                        scope_requested,
                        self.client.client_id,
                    )
                    raise TokenError("invalid_scope")
        # if no scopes requested assign client's scopes
        else:
            token_scopes.extend(self.client.scope)
        return token_scopes

    def create_response_dic(self):
        if self.params["grant_type"] == "authorization_code":
            return self.create_code_response_dic()
        elif self.params["grant_type"] == "refresh_token":
            return self.create_refresh_response_dic()
        elif self.params["grant_type"] == "password":
            return self.create_access_token_response_dic()
        elif self.params["grant_type"] == "client_credentials":
            return self.create_client_credentials_response_dic()

    def create_token(self, user, client, scope):
        token = create_token(
            user=user,
            client=client,
            scope=scope,
        )

        return token

    def create_code_response_dic(self):
        # See https://tools.ietf.org/html/rfc6749#section-4.1

        token = self.create_token(
            user=self.code.user,
            client=self.code.client,
            scope=self.code.scope,
        )

        if self.code.is_authentication:
            id_token_dic = create_id_token(
                user=self.code.user,
                aud=self.client.client_id,
                token=token,
                nonce=self.code.nonce,
                at_hash=token.at_hash,
                request=self.request,
                scope=token.scope,
            )
        else:
            id_token_dic = {}
        token.id_token = id_token_dic

        # Store the token.
        token.save()

        # We don't need to store the code anymore.
        self.code.delete()

        dic = {
            "access_token": token.access_token,
            "refresh_token": token.refresh_token,
            "token_type": "bearer",
            "expires_in": settings.get("OIDC_TOKEN_EXPIRE"),
            "id_token": encode_id_token(id_token_dic, token.client),
        }

        return dic

    def create_refresh_response_dic(self):
        # See https://tools.ietf.org/html/rfc6749#section-6

        scope_param = self.params["scope"]
        scope = scope_param.split(" ") if scope_param else self.token.scope
        unauthorized_scopes = set(scope) - set(self.token.scope)
        if unauthorized_scopes:
            raise TokenError("invalid_scope")

        token = self.create_token(
            user=self.token.user,
            client=self.token.client,
            scope=scope,
        )

        # If the Token has an id_token it's an Authentication request.
        if self.token.id_token:
            id_token_dic = create_id_token(
                user=self.token.user,
                aud=self.client.client_id,
                token=token,
                nonce=None,
                at_hash=token.at_hash,
                request=self.request,
                scope=token.scope,
            )
        else:
            id_token_dic = {}
        token.id_token = id_token_dic

        # Store the token.
        token.save()

        # Forget the old token.
        self.token.delete()

        dic = {
            "access_token": token.access_token,
            "refresh_token": token.refresh_token,
            "token_type": "bearer",
            "expires_in": settings.get("OIDC_TOKEN_EXPIRE"),
            "id_token": encode_id_token(id_token_dic, self.token.client),
        }

        return dic

    def create_access_token_response_dic(self):
        # See https://tools.ietf.org/html/rfc6749#section-4.3
        token_scopes = self.validate_requested_scopes()
        token = self.create_token(
            self.user,
            self.client,
            token_scopes,
        )

        id_token_dic = create_id_token(
            token=token,
            user=self.user,
            aud=self.client.client_id,
            nonce="self.code.nonce",
            at_hash=token.at_hash,
            request=self.request,
            scope=token.scope,
        )

        token.id_token = id_token_dic
        token.save()

        return {
            "access_token": token.access_token,
            "refresh_token": token.refresh_token,
            "expires_in": settings.get("OIDC_TOKEN_EXPIRE"),
            "token_type": "bearer",
            "id_token": encode_id_token(id_token_dic, token.client),
            "scope": " ".join(token.scope),
        }

    def create_client_credentials_response_dic(self):
        # See https://tools.ietf.org/html/rfc6749#section-4.4.3
        token_scopes = self.validate_requested_scopes()

        token = self.create_token(
            user=None,
            client=self.client,
            scope=token_scopes,
        )
        token.save()

        return {
            "access_token": token.access_token,
            "expires_in": settings.get("OIDC_TOKEN_EXPIRE"),
            "token_type": "bearer",
            "scope": " ".join(token.scope),
        }

    @classmethod
    def response(cls, dic, status=200):
        """
        Create and return a response object.
        """
        response = JsonResponse(dic, status=status)
        response["Cache-Control"] = "no-store"
        response["Pragma"] = "no-cache"

        return response



from django.urls import re_path
from django.views.decorators.csrf import csrf_exempt

from oidc_provider import (
    settings,
    views,
)

app_name = "oidc_provider"
urlpatterns = [
    re_path(r"^authorize/?$", views.AuthorizeView.as_view(), name="authorize"),
    re_path(r"^token/?$", csrf_exempt(views.TokenView.as_view()), name="token"),
    re_path(r"^userinfo/?$", csrf_exempt(views.userinfo), name="userinfo"),
    re_path(r"^end-session/?$", views.EndSessionView.as_view(), name="end-session"),
    re_path(
        r"^end-session-prompt/?$", views.EndSessionPromptView.as_view(), name="end-session-prompt"
    ),
    re_path(
        r"^\.well-known/openid-configuration/?$",
        views.ProviderInfoView.as_view(),
        name="provider-info",
    ),
    re_path(r"^introspect/?$", views.TokenIntrospectionView.as_view(), name="token-introspection"),
    re_path(r"^jwks/?$", views.JwksView.as_view(), name="jwks"),
]

if settings.get("OIDC_SESSION_MANAGEMENT_ENABLE"):
    urlpatterns += [
        re_path(
            r"^check-session-iframe/?$",
            views.CheckSessionIframeView.as_view(),
            name="check-session-iframe",
        ),
    ]



# Generated by Django 4.2 on 2024-10-03 19:10

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0026_client_multiple_response_types'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='rsakey',
            options={'ordering': ['id'], 'verbose_name': 'RSA Key', 'verbose_name_plural': 'RSA Keys'},
        ),
    ]
